# Copyright 2015 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Parent client for calling the Google Cloud Bigtable API.

This is the base from which all interactions with the API occur.

In the hierarchy of API concepts

* a :class:`~google.cloud.bigtable.client.Client` owns an
  :class:`~google.cloud.bigtable.instance.Instance`
* an :class:`~google.cloud.bigtable.instance.Instance` owns a
  :class:`~google.cloud.bigtable.table.Table`
* a :class:`~google.cloud.bigtable.table.Table` owns a
  :class:`~.column_family.ColumnFamily`
* a :class:`~google.cloud.bigtable.table.Table` owns a
  :class:`~google.cloud.bigtable.row.Row` (and all the cells in the row)
"""
import os
import warnings
import grpc  # type: ignore

from google.api_core.gapic_v1 import client_info as client_info_lib
from google.auth.credentials import AnonymousCredentials  # type: ignore

from google.cloud import bigtable_v2
from google.cloud import bigtable_admin_v2
from google.cloud.bigtable_v2.services.bigtable.transports import BigtableGrpcTransport
from google.cloud.bigtable_admin_v2.services.bigtable_instance_admin.transports import (
    BigtableInstanceAdminGrpcTransport,
)
from google.cloud.bigtable_admin_v2.services.bigtable_table_admin.transports import (
    BigtableTableAdminGrpcTransport,
)

from google.cloud import bigtable
from google.cloud.bigtable.instance import Instance
from google.cloud.bigtable.cluster import Cluster

from google.cloud.client import ClientWithProject  # type: ignore

from google.cloud.bigtable_admin_v2.types import instance
from google.cloud.bigtable.cluster import _CLUSTER_NAME_RE
from google.cloud.environment_vars import BIGTABLE_EMULATOR  # type: ignore


INSTANCE_TYPE_PRODUCTION = instance.Instance.Type.PRODUCTION
INSTANCE_TYPE_DEVELOPMENT = instance.Instance.Type.DEVELOPMENT
INSTANCE_TYPE_UNSPECIFIED = instance.Instance.Type.TYPE_UNSPECIFIED
SPANNER_ADMIN_SCOPE = "https://www.googleapis.com/auth/spanner.admin"
ADMIN_SCOPE = "https://www.googleapis.com/auth/bigtable.admin"
"""Scope for interacting with the Cluster Admin and Table Admin APIs."""
DATA_SCOPE = "https://www.googleapis.com/auth/bigtable.data"
"""Scope for reading and writing table data."""
READ_ONLY_SCOPE = "https://www.googleapis.com/auth/bigtable.data.readonly"
"""Scope for reading table data."""

_DEFAULT_BIGTABLE_EMULATOR_CLIENT = "google-cloud-bigtable-emulator"
_GRPC_CHANNEL_OPTIONS = (
    ("grpc.max_send_message_length", -1),
    ("grpc.max_receive_message_length", -1),
    ("grpc.keepalive_time_ms", 30000),
    ("grpc.keepalive_timeout_ms", 10000),
)


def _create_gapic_client(client_class, client_options=None, transport=None):
    def inner(self):
        return client_class(
            credentials=None,
            client_info=self._client_info,
            client_options=client_options,
            transport=transport,
        )

    return inner


class Client(ClientWithProject):
    """Client for interacting with Google Cloud Bigtable API.

    .. note::

        Since the Cloud Bigtable API requires the gRPC transport, no
        ``_http`` argument is accepted by this class.

    :type project: :class:`str` or :func:`unicode <unicode>`
    :param project: (Optional) The ID of the project which owns the
                    instances, tables and data. If not provided, will
                    attempt to determine from the environment.

    :type credentials: :class:`~google.auth.credentials.Credentials`
    :param credentials: (Optional) The OAuth2 Credentials to use for this
                        client. If not passed, falls back to the default
                        inferred from the environment.

    :type read_only: bool
    :param read_only: (Optional) Boolean indicating if the data scope should be
                      for reading only (or for writing as well). Defaults to
                      :data:`False`.

    :type admin: bool
    :param admin: (Optional) Boolean indicating if the client will be used to
                  interact with the Instance Admin or Table Admin APIs. This
                  requires the :const:`ADMIN_SCOPE`. Defaults to :data:`False`.

    :type: client_info: :class:`google.api_core.gapic_v1.client_info.ClientInfo`
    :param client_info:
        The client info used to send a user-agent string along with API
        requests. If ``None``, then default info will be used. Generally,
        you only need to set this if you're developing your own library
        or partner tool.

    :type client_options: :class:`~google.api_core.client_options.ClientOptions`
        or :class:`dict`
    :param client_options: (Optional) Client options used to set user options
        on the client. API Endpoint should be set through client_options.

    :type admin_client_options:
        :class:`~google.api_core.client_options.ClientOptions` or :class:`dict`
    :param admin_client_options: (Optional) Client options used to set user
        options on the client. API Endpoint for admin operations should be set
        through admin_client_options.

    :type channel: :instance: grpc.Channel
    :param channel (grpc.Channel): (Optional) DEPRECATED:
            A ``Channel`` instance through which to make calls.
            This argument is mutually exclusive with ``credentials``;
            providing both will raise an exception. No longer used.

    :raises: :class:`ValueError <exceptions.ValueError>` if both ``read_only``
             and ``admin`` are :data:`True`
    """

    _table_data_client = None
    _table_admin_client = None
    _instance_admin_client = None

    def __init__(
        self,
        project=None,
        credentials=None,
        read_only=False,
        admin=False,
        client_info=None,
        client_options=None,
        admin_client_options=None,
        channel=None,
    ):
        if client_info is None:
            client_info = client_info_lib.ClientInfo(
                client_library_version=bigtable.__version__,
            )
        if read_only and admin:
            raise ValueError(
                "A read-only client cannot also perform" "administrative actions."
            )

        # NOTE: We set the scopes **before** calling the parent constructor.
        #       It **may** use those scopes in ``with_scopes_if_required``.
        self._read_only = bool(read_only)
        self._admin = bool(admin)
        self._client_info = client_info
        self._emulator_host = os.getenv(BIGTABLE_EMULATOR)

        if self._emulator_host is not None:
            if credentials is None:
                credentials = AnonymousCredentials()
            if project is None:
                project = _DEFAULT_BIGTABLE_EMULATOR_CLIENT

        if channel is not None:
            warnings.warn(
                "'channel' is deprecated and no longer used.",
                DeprecationWarning,
                stacklevel=2,
            )

        self._client_options = client_options
        self._admin_client_options = admin_client_options
        self._channel = channel
        self.SCOPE = self._get_scopes()
        super(Client, self).__init__(
            project=project,
            credentials=credentials,
            client_options=client_options,
        )

    def _get_scopes(self):
        """Get the scopes corresponding to admin / read-only state.

        Returns:
            Tuple[str, ...]: The tuple of scopes.
        """
        if self._read_only:
            scopes = (READ_ONLY_SCOPE,)
        else:
            scopes = (DATA_SCOPE,)

        if self._admin:
            scopes += (ADMIN_SCOPE,)

        return scopes

    def _emulator_channel(self, transport, options):
        """Create a channel for use with the Bigtable emulator.

        Insecure channels are used for the emulator as secure channels
        cannot be used to communicate on some environments.
        https://github.com/googleapis/python-firestore/issues/359

        Returns:
            grpc.Channel or grpc.aio.Channel
        """
        # Note: this code also exists in the firestore client.
        if "GrpcAsyncIOTransport" in str(transport.__name__):
            channel_fn = grpc.aio.insecure_channel
        else:
            channel_fn = grpc.insecure_channel
        return channel_fn(self._emulator_host, options=options)

    def _create_gapic_client_channel(self, client_class, grpc_transport):
        if self._emulator_host is not None:
            api_endpoint = self._emulator_host
        elif self._client_options and self._client_options.api_endpoint:
            api_endpoint = self._client_options.api_endpoint
        else:
            api_endpoint = client_class.DEFAULT_ENDPOINT

        if self._emulator_host is not None:
            channel = self._emulator_channel(
                transport=grpc_transport,
                options=_GRPC_CHANNEL_OPTIONS,
            )
        else:
            channel = grpc_transport.create_channel(
                host=api_endpoint,
                credentials=self._credentials,
                options=_GRPC_CHANNEL_OPTIONS,
            )
        return grpc_transport(channel=channel, host=api_endpoint)

    @property
    def project_path(self):
        """Project name to be used with Instance Admin API.

        .. note::

            This property will not change if ``project`` does not, but the
            return value is not cached.

        For example:

        .. literalinclude:: snippets.py
            :start-after: [START bigtable_api_project_path]
            :end-before: [END bigtable_api_project_path]
            :dedent: 4

        The project name is of the form

            ``"projects/{project}"``

        :rtype: str
        :returns: Return a fully-qualified project string.
        """
        return self.instance_admin_client.common_project_path(self.project)

    @property
    def table_data_client(self):
        """Getter for the gRPC stub used for the Table Admin API.

        For example:

        .. literalinclude:: snippets.py
            :start-after: [START bigtable_api_table_data_client]
            :end-before: [END bigtable_api_table_data_client]
            :dedent: 4

        :rtype: :class:`.bigtable_v2.BigtableClient`
        :returns: A BigtableClient object.
        """
        if self._table_data_client is None:
            transport = self._create_gapic_client_channel(
                bigtable_v2.BigtableClient,
                BigtableGrpcTransport,
            )
            klass = _create_gapic_client(
                bigtable_v2.BigtableClient,
                client_options=self._client_options,
                transport=transport,
            )
            self._table_data_client = klass(self)
        return self._table_data_client

    @property
    def table_admin_client(self):
        """Getter for the gRPC stub used for the Table Admin API.

        For example:

        .. literalinclude:: snippets.py
            :start-after: [START bigtable_api_table_admin_client]
            :end-before: [END bigtable_api_table_admin_client]
            :dedent: 4

        :rtype: :class:`.bigtable_admin_pb2.BigtableTableAdmin`
        :returns: A BigtableTableAdmin instance.
        :raises: :class:`ValueError <exceptions.ValueError>` if the current
                 client is not an admin client or if it has not been
                 :meth:`start`-ed.
        """
        if self._table_admin_client is None:
            if not self._admin:
                raise ValueError("Client is not an admin client.")

            transport = self._create_gapic_client_channel(
                bigtable_admin_v2.BaseBigtableTableAdminClient,
                BigtableTableAdminGrpcTransport,
            )
            klass = _create_gapic_client(
                bigtable_admin_v2.BaseBigtableTableAdminClient,
                client_options=self._admin_client_options,
                transport=transport,
            )
            self._table_admin_client = klass(self)
        return self._table_admin_client

    @property
    def instance_admin_client(self):
        """Getter for the gRPC stub used for the Table Admin API.

        For example:

        .. literalinclude:: snippets.py
            :start-after: [START bigtable_api_instance_admin_client]
            :end-before: [END bigtable_api_instance_admin_client]
            :dedent: 4

        :rtype: :class:`.bigtable_admin_pb2.BigtableInstanceAdmin`
        :returns: A BigtableInstanceAdmin instance.
        :raises: :class:`ValueError <exceptions.ValueError>` if the current
                 client is not an admin client or if it has not been
                 :meth:`start`-ed.
        """
        if self._instance_admin_client is None:
            if not self._admin:
                raise ValueError("Client is not an admin client.")

            transport = self._create_gapic_client_channel(
                bigtable_admin_v2.BigtableInstanceAdminClient,
                BigtableInstanceAdminGrpcTransport,
            )
            klass = _create_gapic_client(
                bigtable_admin_v2.BigtableInstanceAdminClient,
                client_options=self._admin_client_options,
                transport=transport,
            )
            self._instance_admin_client = klass(self)
        return self._instance_admin_client

    def instance(self, instance_id, display_name=None, instance_type=None, labels=None):
        """Factory to create a instance associated with this client.

        For example:

        .. literalinclude:: snippets.py
            :start-after: [START bigtable_api_create_prod_instance]
            :end-before: [END bigtable_api_create_prod_instance]
            :dedent: 4

        :type instance_id: str
        :param instance_id: The ID of the instance.

        :type display_name: str
        :param display_name: (Optional) The display name for the instance in
                             the Cloud Console UI. (Must be between 4 and 30
                             characters.) If this value is not set in the
                             constructor, will fall back to the instance ID.

        :type instance_type: int
        :param instance_type: (Optional) The type of the instance.
                               Possible values are represented
                               by the following constants:
                               :data:`google.cloud.bigtable.instance.InstanceType.PRODUCTION`.
                               :data:`google.cloud.bigtable.instance.InstanceType.DEVELOPMENT`,
                               Defaults to
                               :data:`google.cloud.bigtable.instance.InstanceType.UNSPECIFIED`.

        :type labels: dict
        :param labels: (Optional) Labels are a flexible and lightweight
                       mechanism for organizing cloud resources into groups
                       that reflect a customer's organizational needs and
                       deployment strategies. They can be used to filter
                       resources and aggregate metrics. Label keys must be
                       between 1 and 63 characters long. Maximum 64 labels can
                       be associated with a given resource. Label values must
                       be between 0 and 63 characters long. Keys and values
                       must both be under 128 bytes.

        :rtype: :class:`~google.cloud.bigtable.instance.Instance`
        :returns: an instance owned by this client.
        """
        return Instance(
            instance_id,
            self,
            display_name=display_name,
            instance_type=instance_type,
            labels=labels,
        )

    def list_instances(self):
        """List instances owned by the project.

        For example:

        .. literalinclude:: snippets.py
            :start-after: [START bigtable_api_list_instances]
            :end-before: [END bigtable_api_list_instances]
            :dedent: 4

        :rtype: tuple
        :returns:
            (instances, failed_locations), where 'instances' is list of
            :class:`google.cloud.bigtable.instance.Instance`, and
            'failed_locations' is a list of locations which could not
            be resolved.
        """
        resp = self.instance_admin_client.list_instances(
            request={"parent": self.project_path}
        )
        instances = [Instance.from_pb(instance, self) for instance in resp.instances]
        return instances, resp.failed_locations

    def list_clusters(self):
        """List the clusters in the project.

        For example:

        .. literalinclude:: snippets.py
            :start-after: [START bigtable_api_list_clusters_in_project]
            :end-before: [END bigtable_api_list_clusters_in_project]
            :dedent: 4

        :rtype: tuple
        :returns:
            (clusters, failed_locations), where 'clusters' is list of
            :class:`google.cloud.bigtable.instance.Cluster`, and
            'failed_locations' is a list of strings representing
            locations which could not be resolved.
        """
        resp = self.instance_admin_client.list_clusters(
            request={
                "parent": self.instance_admin_client.instance_path(self.project, "-")
            }
        )
        clusters = []
        instances = {}
        for cluster in resp.clusters:
            match_cluster_name = _CLUSTER_NAME_RE.match(cluster.name)
            instance_id = match_cluster_name.group("instance")
            if instance_id not in instances:
                instances[instance_id] = self.instance(instance_id)
            clusters.append(Cluster.from_pb(cluster, instances[instance_id]))
        return clusters, resp.failed_locations



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Factory for creating MetricTracer instances, facilitating metrics collection and tracing."""

from google.cloud.spanner_v1.metrics.metrics_tracer import MetricsTracer

from google.cloud.spanner_v1.metrics.constants import (
    METRIC_NAME_OPERATION_LATENCIES,
    MONITORED_RES_LABEL_KEY_PROJECT,
    METRIC_NAME_ATTEMPT_LATENCIES,
    METRIC_NAME_OPERATION_COUNT,
    METRIC_NAME_ATTEMPT_COUNT,
    MONITORED_RES_LABEL_KEY_INSTANCE,
    MONITORED_RES_LABEL_KEY_INSTANCE_CONFIG,
    MONITORED_RES_LABEL_KEY_LOCATION,
    MONITORED_RES_LABEL_KEY_CLIENT_HASH,
    METRIC_LABEL_KEY_CLIENT_UID,
    METRIC_LABEL_KEY_CLIENT_NAME,
    METRIC_LABEL_KEY_DATABASE,
    METRIC_LABEL_KEY_DIRECT_PATH_ENABLED,
    BUILT_IN_METRICS_METER_NAME,
    METRIC_NAME_GFE_LATENCY,
    METRIC_NAME_GFE_MISSING_HEADER_COUNT,
)

from typing import Dict

try:
    from opentelemetry.metrics import Counter, Histogram, get_meter_provider

    HAS_OPENTELEMETRY_INSTALLED = True
except ImportError:  # pragma: NO COVER
    HAS_OPENTELEMETRY_INSTALLED = False

from google.cloud.spanner_v1 import __version__


class MetricsTracerFactory:
    """Factory class for creating MetricTracer instances. This class facilitates the creation of MetricTracer objects, which are responsible for collecting and tracing metrics."""

    enabled: bool
    gfe_enabled: bool
    _instrument_attempt_latency: "Histogram"
    _instrument_attempt_counter: "Counter"
    _instrument_operation_latency: "Histogram"
    _instrument_operation_counter: "Counter"
    _instrument_gfe_latency: "Histogram"
    _instrument_gfe_missing_header_count: "Counter"
    _client_attributes: Dict[str, str]

    @property
    def instrument_attempt_latency(self) -> "Histogram":
        return self._instrument_attempt_latency

    @property
    def instrument_attempt_counter(self) -> "Counter":
        return self._instrument_attempt_counter

    @property
    def instrument_operation_latency(self) -> "Histogram":
        return self._instrument_operation_latency

    @property
    def instrument_operation_counter(self) -> "Counter":
        return self._instrument_operation_counter

    def __init__(self, enabled: bool, service_name: str):
        """Initialize a MetricsTracerFactory instance with the given parameters.

        This constructor initializes a MetricsTracerFactory instance with the provided service name, project, instance, instance configuration, location, client hash, client UID, client name, and database. It sets up the necessary metric instruments and client attributes for metrics tracing.

        Args:
            service_name (str): The name of the service for which metrics are being traced.
            project (str): The project ID for the monitored resource.
        """
        self.enabled = enabled
        self._create_metric_instruments(service_name)
        self._client_attributes = {}

    @property
    def client_attributes(self) -> Dict[str, str]:
        """Return a dictionary of client attributes used for metrics tracing.

        This property returns a dictionary containing client attributes such as project, instance,
        instance configuration, location, client hash, client UID, client name, and database.
        These attributes are used to provide context to the metrics being traced.

        Returns:
            dict[str, str]: A dictionary of client attributes.
        """
        return self._client_attributes

    def set_project(self, project: str) -> "MetricsTracerFactory":
        """Set the project attribute for metrics tracing.

        This method updates the client attributes dictionary with the provided project name.
        The project name is used to identify the project for which metrics are being traced
        and is passed to the created MetricsTracer.

        Args:
            project (str): The name of the project for metrics tracing.

        Returns:
            MetricsTracerFactory: The current instance of MetricsTracerFactory to enable method chaining.
        """
        self._client_attributes[MONITORED_RES_LABEL_KEY_PROJECT] = project
        return self

    def set_instance(self, instance: str) -> "MetricsTracerFactory":
        """Set the instance attribute for metrics tracing.

        This method updates the client attributes dictionary with the provided instance name.
        The instance name is used to identify the instance for which metrics are being traced
        and is passed to the created MetricsTracer.

        Args:
            instance (str): The name of the instance for metrics tracing.

        Returns:
            MetricsTracerFactory: The current instance of MetricsTracerFactory to enable method chaining.
        """
        self._client_attributes[MONITORED_RES_LABEL_KEY_INSTANCE] = instance
        return self

    def set_instance_config(self, instance_config: str) -> "MetricsTracerFactory":
        """Sets the instance configuration attribute for metrics tracing.

        This method updates the client attributes dictionary with the provided instance configuration.
        The instance configuration is used to identify the configuration of the instance for which
        metrics are being traced and is passed to the created MetricsTracer.

        Args:
            instance_config (str): The configuration of the instance for metrics tracing.

        Returns:
            MetricsTracerFactory: The current instance of MetricsTracerFactory to enable method chaining.
        """
        self._client_attributes[
            MONITORED_RES_LABEL_KEY_INSTANCE_CONFIG
        ] = instance_config
        return self

    def set_location(self, location: str) -> "MetricsTracerFactory":
        """Set the location attribute for metrics tracing.

        This method updates the client attributes dictionary with the provided location.
        The location is used to identify the location for which metrics are being traced
        and is passed to the created MetricsTracer.

        Args:
            location (str): The location for metrics tracing.

        Returns:
            MetricsTracerFactory: The current instance of MetricsTracerFactory to enable method chaining.
        """
        self._client_attributes[MONITORED_RES_LABEL_KEY_LOCATION] = location
        return self

    def set_client_hash(self, hash: str) -> "MetricsTracerFactory":
        """Set the client hash attribute for metrics tracing.

        This method updates the client attributes dictionary with the provided client hash.
        The client hash is used to identify the client for which metrics are being traced
        and is passed to the created MetricsTracer.

        Args:
            hash (str): The hash of the client for metrics tracing.

        Returns:
            MetricsTracerFactory: The current instance of MetricsTracerFactory to enable method chaining.
        """
        self._client_attributes[MONITORED_RES_LABEL_KEY_CLIENT_HASH] = hash
        return self

    def set_client_uid(self, client_uid: str) -> "MetricsTracerFactory":
        """Set the client UID attribute for metrics tracing.

        This method updates the client attributes dictionary with the provided client UID.
        The client UID is used to identify the client for which metrics are being traced
        and is passed to the created MetricsTracer.

        Args:
            client_uid (str): The UID of the client for metrics tracing.

        Returns:
            MetricsTracerFactory: The current instance of MetricsTracerFactory to enable method chaining.
        """
        self._client_attributes[METRIC_LABEL_KEY_CLIENT_UID] = client_uid
        return self

    def set_client_name(self, client_name: str) -> "MetricsTracerFactory":
        """Set the client name attribute for metrics tracing.

        This method updates the client attributes dictionary with the provided client name.
        The client name is used to identify the client for which metrics are being traced
        and is passed to the created MetricsTracer.

        Args:
            client_name (str): The name of the client for metrics tracing.

        Returns:
            MetricsTracerFactory: The current instance of MetricsTracerFactory to enable method chaining.
        """
        self._client_attributes[METRIC_LABEL_KEY_CLIENT_NAME] = client_name
        return self

    def set_database(self, database: str) -> "MetricsTracerFactory":
        """Set the database attribute for metrics tracing.

        This method updates the client attributes dictionary with the provided database name.
        The database name is used to identify the database for which metrics are being traced
        and is passed to the created MetricsTracer.

        Args:
            database (str): The name of the database for metrics tracing.

        Returns:
            MetricsTracerFactory: The current instance of MetricsTracerFactory to enable method chaining.
        """
        self._client_attributes[METRIC_LABEL_KEY_DATABASE] = database
        return self

    def enable_direct_path(self, enable: bool = False) -> "MetricsTracerFactory":
        """Enable or disable the direct path for metrics tracing.

        This method updates the client attributes dictionary with the provided enable status.
        The direct path enabled status is used to determine whether to use the direct path for metrics tracing
        and is passed to the created MetricsTracer.

        Args:
            enable (bool, optional): Whether to enable the direct path for metrics tracing. Defaults to False.

        Returns:
            MetricsTracerFactory: The current instance of MetricsTracerFactory to enable method chaining.
        """
        self._client_attributes[METRIC_LABEL_KEY_DIRECT_PATH_ENABLED] = enable
        return self

    def create_metrics_tracer(self) -> MetricsTracer:
        """
        Create and return a MetricsTracer instance with default settings and client attributes.

        This method initializes a MetricsTracer instance with default settings for metrics tracing,
        including metrics tracing enabled if OpenTelemetry is installed and the direct path disabled by default.
        It also sets the client attributes based on the factory's configuration.

        Returns:
            MetricsTracer: A MetricsTracer instance with default settings and client attributes.
        """
        if not HAS_OPENTELEMETRY_INSTALLED:
            return None

        metrics_tracer = MetricsTracer(
            enabled=self.enabled and HAS_OPENTELEMETRY_INSTALLED,
            instrument_attempt_latency=self._instrument_attempt_latency,
            instrument_attempt_counter=self._instrument_attempt_counter,
            instrument_operation_latency=self._instrument_operation_latency,
            instrument_operation_counter=self._instrument_operation_counter,
            client_attributes=self._client_attributes.copy(),
        )
        return metrics_tracer

    def _create_metric_instruments(self, service_name: str) -> None:
        """
        Creates and sets up metric instruments for the given service name.

        This method initializes and configures metric instruments for attempt latency, attempt counter,
        operation latency, and operation counter. These instruments are used to measure and track
        metrics related to attempts and operations within the service.

        Args:
            service_name (str): The name of the service for which metric instruments are being created.
        """
        if not HAS_OPENTELEMETRY_INSTALLED:  # pragma: NO COVER
            return

        meter_provider = get_meter_provider()
        meter = meter_provider.get_meter(
            name=BUILT_IN_METRICS_METER_NAME, version=__version__
        )

        self._instrument_attempt_latency = meter.create_histogram(
            name=METRIC_NAME_ATTEMPT_LATENCIES,
            unit="ms",
            description="Time an individual attempt took.",
        )

        self._instrument_attempt_counter = meter.create_counter(
            name=METRIC_NAME_ATTEMPT_COUNT,
            unit="1",
            description="Number of attempts.",
        )

        self._instrument_operation_latency = meter.create_histogram(
            name=METRIC_NAME_OPERATION_LATENCIES,
            unit="ms",
            description="Total time until final operation success or failure, including retries and backoff.",
        )

        self._instrument_operation_counter = meter.create_counter(
            name=METRIC_NAME_OPERATION_COUNT,
            unit="1",
            description="Number of operations.",
        )

        self._instrument_gfe_latency = meter.create_histogram(
            name=METRIC_NAME_GFE_LATENCY,
            unit="ms",
            description="GFE Latency.",
        )

        self._instrument_gfe_missing_header_count = meter.create_counter(
            name=METRIC_NAME_GFE_MISSING_HEADER_COUNT,
            unit="1",
            description="GFE missing header count.",
        )



#!/usr/bin/env python3
"""
Ultra-Minimal BMasterAI RAG - In-Memory Implementation
No external vector DB required - uses bmasterai + numpy
"""

import os
import gradio as gr
import numpy as np
from bmasterai.logging import configure_logging, get_logger, EventType
from bmasterai.monitoring import get_monitor
from sentence_transformers import SentenceTransformer
import requests

# BMasterAI setup
logger = configure_logging()
monitor = get_monitor()

class UltraMinimalRAG:
    def __init__(self):
        self.embedder = SentenceTransformer('all-MiniLM-L6-v2')
        self.vectors = []  # In-memory storage
        self.texts = []
        self.sources = []
        self.api_key = os.getenv("ANTHROPIC_API_KEY", "")
        logger.log_event("system", EventType.AGENT_START, "Ultra-minimal RAG started")
    
    def upload(self, file):
        if not file: return "No file"
        
        # Read text
        text = file.read().decode('utf-8') if file.name.endswith('.txt') else "PDF support not included"
        
        # Simple chunking
        chunks = [text[i:i+500] for i in range(0, len(text), 400)]
        
        # Store embeddings
        for chunk in chunks:
            self.vectors.append(self.embedder.encode(chunk))
            self.texts.append(chunk)
            self.sources.append(file.name)
        
        monitor.record_metric("docs_uploaded", 1)
        return f"✅ Uploaded {file.name} ({len(chunks)} chunks)"
    
    def query(self, question):
        if not question or not self.vectors:
            return "No question or documents"
        
        # Find similar chunks
        q_vec = self.embedder.encode(question)
        scores = [np.dot(q_vec, v) / (np.linalg.norm(q_vec) * np.linalg.norm(v)) 
                  for v in self.vectors]
        top_idx = sorted(range(len(scores)), key=lambda i: scores[i], reverse=True)[:3]
        
        # Build context
        context = "\n".join([f"[{self.sources[i]}]: {self.texts[i]}" for i in top_idx])
        
        # Ask Claude
        resp = requests.post(
            "https://api.anthropic.com/v1/messages",
            headers={"x-api-key": self.api_key, "anthropic-version": "2023-06-01"},
            json={
                "model": "claude-3-5-sonnet-20241022",
                "max_tokens": 500,
                "messages": [{"role": "user", "content": f"Context:\n{context}\n\nQ: {question}"}]
            }
        )
        
        monitor.record_metric("queries", 1)
        return resp.json()["content"][0]["text"] if resp.ok else "Error"

# Gradio UI
def create_ui():
    rag = UltraMinimalRAG()
    
    return gr.Interface(
        fn=lambda file, q, action: rag.upload(file) if action == "upload" else rag.query(q),
        inputs=[
            gr.File(label="Upload .txt file"),
            gr.Textbox(label="Question"),
            gr.Radio(["upload", "query"], label="Action")
        ],
        outputs=gr.Textbox(label="Result"),
        title="🚀 Ultra-Minimal BMasterAI RAG",
        description="Simplest possible RAG with BMasterAI logging/monitoring"
    )

if __name__ == "__main__":
    create_ui().launch()



# -*- coding: utf-8 -*-

# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Helper functions for processing and formatting citations from RAG generation outputs."""


def populate_cited_chunk_references(
    grounding_supports,
    grounding_chunks,
    cited_refs_dict,
) -> None:
    """Populates cited_refs_dict with URI information for all unique chunk indices found in grounding_supports.

    Args:
        grounding_supports: A list of support items, where each item might contain
          grounding chunk indices.
        grounding_chunks: A list of all available chunk items from which to retrieve
          context and URI.
        cited_refs_dict: A dictionary to populate with chunk_idx as key and URI
          as value.

    Raises:
        TypeError: If grounding_chunks is not a list, or cited_refs_dict is not a
          dictionary.
        ValueError: If grounding_chunks or cited_refs_dict is None.
                    If a chunk_item at a valid index is None.
                    If 'retrieved_context' or 'uri' attribute of a chunk is None.
        IndexError: If a chunk_idx is out of bounds for grounding_chunks.
        AttributeError: If 'retrieved_context' or 'uri' attribute is missing from
          a chunk or its context.
    """
    if grounding_chunks is None or not grounding_chunks:
        raise ValueError("grounding_chunks cannot be None or empty.")
    if grounding_supports is None or not grounding_supports:
        raise ValueError("grounding_supports cannot be None or empty.")
    if not isinstance(grounding_chunks, list):
        raise TypeError("grounding_chunks must be a list.")
    if not isinstance(grounding_supports, list):
        raise TypeError("grounding_supports must be a list.")
    if cited_refs_dict is None:
        raise ValueError("cited_refs_dict cannot be None.")
    if not isinstance(cited_refs_dict, dict):
        raise TypeError("cited_refs_dict must be a dictionary.")

    for support in grounding_supports:
        current_support_chunk_indices = []
        if (
            hasattr(support, "grounding_chunk_indices")
            and support.grounding_chunk_indices is not None
        ):
            valid_indices = [
                idx for idx in support.grounding_chunk_indices if isinstance(idx, int)
            ]
            current_support_chunk_indices = sorted(list(set(valid_indices)))

        for chunk_idx in current_support_chunk_indices:
            if chunk_idx not in cited_refs_dict:
                if not (0 <= chunk_idx < len(grounding_chunks)):
                    raise IndexError(
                        f"Chunk index {chunk_idx} is out of bounds for grounding_chunks of size {len(grounding_chunks)}."
                    )
                chunk_item = grounding_chunks[chunk_idx]
                if chunk_item is None:
                    raise ValueError(f"Chunk item at index {chunk_idx} is None.")
                if not hasattr(chunk_item, "retrieved_context"):
                    raise AttributeError(
                        f"Chunk item at index {chunk_idx} is missing 'retrieved_context' attribute."
                    )
                retrieved_context_obj = chunk_item.retrieved_context
                if retrieved_context_obj is None:
                    raise ValueError(
                        f"Attribute 'retrieved_context' for chunk {chunk_idx} is None."
                    )
                if not hasattr(retrieved_context_obj, "uri"):
                    raise AttributeError(
                        f"retrieved_context for chunk {chunk_idx} is missing 'uri' attribute."
                    )
                uri = retrieved_context_obj.uri
                if uri is None:
                    raise ValueError(f"Attribute 'uri' for chunk {chunk_idx} is None.")
                cited_refs_dict[chunk_idx] = uri


def format_bibliography(cited_refs_dict, grounding_chunks) -> str:
    """Formats the bibliography string from the populated cited_refs_dict.

    Omits page information if page numbers are not valid (e.g., not >= 1).

    Args:
        cited_refs_dict: A dictionary with chunk_idx as key and URI as value.
          It's expected that populate_cited_chunk_references has successfully
          populated this dict.
        grounding_chunks: A list of all available chunk items, used to retrieve page
          span information.

    Returns:
        A string representing the formatted bibliography, with each reference
        on a new line.

    Raises:
        TypeError: If cited_refs_dict is not a dictionary or grounding_chunks is not a list.
        ValueError: If cited_refs_dict or grounding_chunks is None.
                    If a chunk_item in grounding_chunks referenced by cited_refs_dict is None.
        IndexError: If a chunk_idx from cited_refs_dict is out of bounds for
          grounding_chunks.
    """
    if cited_refs_dict is None:
        raise ValueError("cited_refs_dict cannot be None.")
    if not isinstance(cited_refs_dict, dict):
        raise TypeError("cited_refs_dict must be a dictionary.")
    if grounding_chunks is None:
        raise ValueError("grounding_chunks cannot be None.")
    if not isinstance(grounding_chunks, list):
        raise TypeError("grounding_chunks must be a list.")

    reference_lines = []
    for chunk_idx_ref in sorted(list(cited_refs_dict.keys())):
        uri = cited_refs_dict[chunk_idx_ref]
        page_info_str = ""
        if not (
            isinstance(chunk_idx_ref, int)
            and 0 <= chunk_idx_ref < len(grounding_chunks)
        ):
            raise IndexError(
                f"Chunk index {chunk_idx_ref} from cited_refs_dict is invalid or out of bounds "
                f"for grounding_chunks of size {len(grounding_chunks)}."
            )
        chunk_item = grounding_chunks[chunk_idx_ref]
        if chunk_item is None:
            raise ValueError(
                f"Chunk item at index {chunk_idx_ref} in grounding_chunks is None, "
                "but was referenced in cited_refs_dict."
            )
        page_span_data = None
        if (
            hasattr(chunk_item, "retrieved_context")
            and chunk_item.retrieved_context
            and hasattr(chunk_item.retrieved_context, "rag_chunk")
            and chunk_item.retrieved_context.rag_chunk
            and hasattr(chunk_item.retrieved_context.rag_chunk, "page_span")
            and chunk_item.retrieved_context.rag_chunk.page_span
        ):
            page_span_data = chunk_item.retrieved_context.rag_chunk.page_span
        if (
            page_span_data
            and hasattr(page_span_data, "first_page")
            and hasattr(page_span_data, "last_page")
        ):
            first_page_val = page_span_data.first_page
            last_page_val = page_span_data.last_page
            is_first_page_valid_num = (
                isinstance(first_page_val, int) and first_page_val >= 1
            )
            is_last_page_valid_num = (
                isinstance(last_page_val, int) and last_page_val >= 1
            )
            if is_first_page_valid_num and is_last_page_valid_num:
                if last_page_val >= first_page_val:
                    page_info_str = (
                        f", p.{first_page_val}-{last_page_val}"
                        if first_page_val != last_page_val
                        else f", p.{first_page_val}"
                    )
        reference_lines.append(f"[{chunk_idx_ref}] {uri}{page_info_str}")
    return "\n".join(reference_lines)



# -*- coding: utf-8 -*-
# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

from typing import MutableMapping, MutableSequence

from google.protobuf import timestamp_pb2  # type: ignore
import proto  # type: ignore

__protobuf__ = proto.module(
    package="google.cloud.resourcemanager.v3",
    manifest={
        "TagHold",
        "CreateTagHoldRequest",
        "CreateTagHoldMetadata",
        "DeleteTagHoldRequest",
        "DeleteTagHoldMetadata",
        "ListTagHoldsRequest",
        "ListTagHoldsResponse",
    },
)


class TagHold(proto.Message):
    r"""A TagHold represents the use of a TagValue that is not captured by
    TagBindings. If a TagValue has any TagHolds, deletion will be
    blocked. This resource is intended to be created in the same cloud
    location as the ``holder``.

    Attributes:
        name (str):
            Output only. The resource name of a TagHold. This is a
            String of the form:
            ``tagValues/{tag-value-id}/tagHolds/{tag-hold-id}`` (e.g.
            ``tagValues/123/tagHolds/456``). This resource name is
            generated by the server.
        holder (str):
            Required. The name of the resource where the TagValue is
            being used. Must be less than 200 characters. E.g.
            ``//compute.googleapis.com/compute/projects/myproject/regions/us-east-1/instanceGroupManagers/instance-group``
        origin (str):
            Optional. An optional string representing the origin of this
            request. This field should include human-understandable
            information to distinguish origins from each other. Must be
            less than 200 characters. E.g. ``migs-35678234``
        help_link (str):
            Optional. A URL where an end user can learn more about
            removing this hold. E.g.
            ``https://cloud.google.com/resource-manager/docs/tags/tags-creating-and-managing``
        create_time (google.protobuf.timestamp_pb2.Timestamp):
            Output only. The time this TagHold was
            created.
    """

    name: str = proto.Field(
        proto.STRING,
        number=1,
    )
    holder: str = proto.Field(
        proto.STRING,
        number=2,
    )
    origin: str = proto.Field(
        proto.STRING,
        number=3,
    )
    help_link: str = proto.Field(
        proto.STRING,
        number=4,
    )
    create_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=5,
        message=timestamp_pb2.Timestamp,
    )


class CreateTagHoldRequest(proto.Message):
    r"""The request message to create a TagHold.

    Attributes:
        parent (str):
            Required. The resource name of the TagHold's parent
            TagValue. Must be of the form: ``tagValues/{tag-value-id}``.
        tag_hold (google.cloud.resourcemanager_v3.types.TagHold):
            Required. The TagHold to be created.
        validate_only (bool):
            Optional. Set to true to perform the
            validations necessary for creating the resource,
            but not actually perform the action.
    """

    parent: str = proto.Field(
        proto.STRING,
        number=1,
    )
    tag_hold: "TagHold" = proto.Field(
        proto.MESSAGE,
        number=2,
        message="TagHold",
    )
    validate_only: bool = proto.Field(
        proto.BOOL,
        number=3,
    )


class CreateTagHoldMetadata(proto.Message):
    r"""Runtime operation information for creating a TagHold.
    (-- The metadata is currently empty, but may include information
    in the future. --)

    """


class DeleteTagHoldRequest(proto.Message):
    r"""The request message to delete a TagHold.

    Attributes:
        name (str):
            Required. The resource name of the TagHold to delete. Must
            be of the form:
            ``tagValues/{tag-value-id}/tagHolds/{tag-hold-id}``.
        validate_only (bool):
            Optional. Set to true to perform the
            validations necessary for deleting the resource,
            but not actually perform the action.
    """

    name: str = proto.Field(
        proto.STRING,
        number=1,
    )
    validate_only: bool = proto.Field(
        proto.BOOL,
        number=2,
    )


class DeleteTagHoldMetadata(proto.Message):
    r"""Runtime operation information for deleting a TagHold.
    (-- The metadata is currently empty, but may include information
    in the future. --)

    """


class ListTagHoldsRequest(proto.Message):
    r"""The request message for listing the TagHolds under a
    TagValue.

    Attributes:
        parent (str):
            Required. The resource name of the parent TagValue. Must be
            of the form: ``tagValues/{tag-value-id}``.
        page_size (int):
            Optional. The maximum number of TagHolds to
            return in the response. The server allows a
            maximum of 300 TagHolds to return. If
            unspecified, the server will use 100 as the
            default.
        page_token (str):
            Optional. A pagination token returned from a previous call
            to ``ListTagHolds`` that indicates where this listing should
            continue from.
        filter (str):
            Optional. Criteria used to select a subset of TagHolds
            parented by the TagValue to return. This field follows the
            syntax defined by aip.dev/160; the ``holder`` and ``origin``
            fields are supported for filtering. Currently only ``AND``
            syntax is supported. Some example queries are:

            - ``holder = //compute.googleapis.com/compute/projects/myproject/regions/us-east-1/instanceGroupManagers/instance-group``
            - ``origin = 35678234``
            - ``holder = //compute.googleapis.com/compute/projects/myproject/regions/us-east-1/instanceGroupManagers/instance-group AND origin = 35678234``
    """

    parent: str = proto.Field(
        proto.STRING,
        number=1,
    )
    page_size: int = proto.Field(
        proto.INT32,
        number=2,
    )
    page_token: str = proto.Field(
        proto.STRING,
        number=3,
    )
    filter: str = proto.Field(
        proto.STRING,
        number=4,
    )


class ListTagHoldsResponse(proto.Message):
    r"""The ListTagHolds response.

    Attributes:
        tag_holds (MutableSequence[google.cloud.resourcemanager_v3.types.TagHold]):
            A possibly paginated list of TagHolds.
        next_page_token (str):
            Pagination token.

            If the result set is too large to fit in a single response,
            this token is returned. It encodes the position of the
            current result cursor. Feeding this value into a new list
            request with the ``page_token`` parameter gives the next
            page of the results.

            When ``next_page_token`` is not filled in, there is no next
            page and the list returned is the last page in the result
            set.

            Pagination tokens have a limited lifetime.
    """

    @property
    def raw_page(self):
        return self

    tag_holds: MutableSequence["TagHold"] = proto.RepeatedField(
        proto.MESSAGE,
        number=1,
        message="TagHold",
    )
    next_page_token: str = proto.Field(
        proto.STRING,
        number=2,
    )


__all__ = tuple(sorted(__protobuf__.manifest))



# -*- coding: utf-8 -*-

# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Model-based Pairwise Metric."""

from typing import Callable, Optional, Union

from google.cloud.aiplatform_v1beta1.types import (
    evaluation_service as gapic_eval_service_types,
)
from vertexai.preview import generative_models
from vertexai.preview.evaluation.metrics import _base
from vertexai.preview.evaluation.metrics import (
    custom_output_config as custom_output_config_class,
)
from vertexai.preview.evaluation.metrics import (
    metric_prompt_template as metric_prompt_template_base,
)


class PairwiseMetric(_base._ModelBasedMetric):  # pylint: disable=protected-access
    """A Model-based Pairwise Metric.

    A model-based evaluation metric that compares two generative models' responses
    side-by-side, and allows users to A/B test their generative models to
    determine which model is performing better.

    For more details on when to use pairwise metrics, see
    [Evaluation methods and
    metrics](https://cloud.google.com/vertex-ai/generative-ai/docs/models/determine-eval#pointwise_versus_pairwise).

    Result Details:

        * In `EvalResult.summary_metrics`, win rates for both the baseline and
        candidate model are computed. The win rate is computed as proportion of
        wins of one model's responses to total attempts as a decimal value
        between 0 and 1.

        * In `EvalResult.metrics_table`, a pairwise metric produces two
        evaluation results per dataset row:
            * `pairwise_choice`: The choice shows whether the candidate model or
              the baseline model performs better, or if they are equally good.
            * `explanation`: The rationale behind each verdict using
              chain-of-thought reasoning. The explanation helps users scrutinize
              the judgment and builds appropriate trust in the decisions.

        See [documentation
        page](https://cloud.google.com/vertex-ai/generative-ai/docs/models/determine-eval#understand-results)
        for more details on understanding the metric results.

    Usage Examples:

        ```
        baseline_model = GenerativeModel("gemini-1.0-pro")
        candidate_model = GenerativeModel("gemini-1.5-pro")

        pairwise_groundedness = PairwiseMetric(
            metric_prompt_template=MetricPromptTemplateExamples.get_prompt_template(
                "pairwise_groundedness"
            ),
            baseline_model=baseline_model,
        )
        eval_dataset = pd.DataFrame({
              "prompt"  : [...],
        })
        pairwise_task = EvalTask(
            dataset=eval_dataset,
            metrics=[pairwise_groundedness],
            experiment="my-pairwise-experiment",
        )
        pairwise_result = pairwise_task.evaluate(
            model=candidate_model,
            experiment_run_name="gemini-pairwise-eval-run",
        )
        ```
    """

    def __init__(
        self,
        *,
        metric: str,
        metric_prompt_template: Union[
            metric_prompt_template_base.PairwiseMetricPromptTemplate, str
        ],
        baseline_model: Optional[
            Union[generative_models.GenerativeModel, Callable[[str], str]]
        ] = None,
        system_instruction: Optional[str] = None,
        autorater_config: Optional[gapic_eval_service_types.AutoraterConfig] = None,
        custom_output_config: Optional[
            custom_output_config_class.CustomOutputConfig
        ] = None,
    ):
        """Initializes a pairwise evaluation metric.

        Args:
          metric: The pairwise evaluation metric name.
          metric_prompt_template: Pairwise metric prompt template for performing
            the pairwise model-based evaluation. A freeform string is also accepted.
          baseline_model: The baseline model for side-by-side comparison. If not
            specified, `baseline_model_response` column is required in the dataset
            to perform bring-your-own-response(BYOR) evaluation.
          system_instruction: The system instruction for the evaluation.
          autorater_config: The config for judge model.
          custom_output_config: Config for custom output from the judge model.
        """
        super().__init__(
            metric_prompt_template=metric_prompt_template,
            metric=metric,
            system_instruction=system_instruction,
            autorater_config=autorater_config,
            custom_output_config=custom_output_config,
        )
        self._baseline_model = baseline_model

    @property
    def baseline_model(
        self,
    ) -> Union[generative_models.GenerativeModel, Callable[[str], str]]:
        return self._baseline_model



# -*- coding: utf-8 -*-
# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from collections import OrderedDict
from typing import Dict, Type, Tuple

from .base import ExampleStoreServiceTransport
from .grpc import ExampleStoreServiceGrpcTransport
from .grpc_asyncio import ExampleStoreServiceGrpcAsyncIOTransport
from .rest import ExampleStoreServiceRestTransport
from .rest import ExampleStoreServiceRestInterceptor

ASYNC_REST_CLASSES: Tuple[str, ...]
try:
    from .rest_asyncio import AsyncExampleStoreServiceRestTransport
    from .rest_asyncio import AsyncExampleStoreServiceRestInterceptor

    ASYNC_REST_CLASSES = (
        "AsyncExampleStoreServiceRestTransport",
        "AsyncExampleStoreServiceRestInterceptor",
    )
    HAS_REST_ASYNC = True
except ImportError:  # pragma: NO COVER
    ASYNC_REST_CLASSES = ()
    HAS_REST_ASYNC = False


# Compile a registry of transports.
_transport_registry = (
    OrderedDict()
)  # type: Dict[str, Type[ExampleStoreServiceTransport]]
_transport_registry["grpc"] = ExampleStoreServiceGrpcTransport
_transport_registry["grpc_asyncio"] = ExampleStoreServiceGrpcAsyncIOTransport
_transport_registry["rest"] = ExampleStoreServiceRestTransport
if HAS_REST_ASYNC:  # pragma: NO COVER
    _transport_registry["rest_asyncio"] = AsyncExampleStoreServiceRestTransport

__all__ = (
    "ExampleStoreServiceTransport",
    "ExampleStoreServiceGrpcTransport",
    "ExampleStoreServiceGrpcAsyncIOTransport",
    "ExampleStoreServiceRestTransport",
    "ExampleStoreServiceRestInterceptor",
) + ASYNC_REST_CLASSES




import requests
import json
import smtplib
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from typing import Dict, Any, List, Optional
import sqlite3
import os
from datetime import datetime, timezone
import base64
from abc import ABC, abstractmethod

class BaseConnector(ABC):
    def __init__(self, config: Dict[str, Any]):
        self.config = config
        self.name = self.__class__.__name__

    @abstractmethod
    def connect(self) -> bool:
        pass

    @abstractmethod
    def test_connection(self) -> Dict[str, Any]:
        pass

class SlackConnector(BaseConnector):
    """Slack integration for sending messages and notifications"""

    def __init__(self, webhook_url: str, bot_token: Optional[str] = None):
        super().__init__({'webhook_url': webhook_url, 'bot_token': bot_token})
        self.webhook_url = webhook_url
        self.bot_token = bot_token
        self.headers = {
            'Content-Type': 'application/json',
            'Authorization': f'Bearer {bot_token}' if bot_token else None
        }

    def connect(self) -> bool:
        return self.test_connection()['success']

    def test_connection(self) -> Dict[str, Any]:
        try:
            response = self.send_message("🤖 BMasterAI connection test", channel="#general")
            return {'success': True, 'message': 'Connected to Slack successfully'}
        except Exception as e:
            return {'success': False, 'error': str(e)}

    def send_message(self, message: str, channel: str = "#general", username: str = "BMasterAI") -> Dict[str, Any]:
        payload = {
            'text': message,
            'channel': channel,
            'username': username,
            'icon_emoji': ':robot_face:'
        }

        try:
            response = requests.post(self.webhook_url, json=payload, timeout=10)
            response.raise_for_status()
            return {'success': True, 'response': response.text}
        except Exception as e:
            return {'success': False, 'error': str(e)}

    def send_alert(self, alert_data: Dict[str, Any]) -> Dict[str, Any]:
        message = f"🚨 *Alert: {alert_data.get('metric_name', 'Unknown')}*\n"
        message += f"Current Value: {alert_data.get('current_value', 'N/A')}\n"
        message += f"Threshold: {alert_data.get('threshold', 'N/A')}\n"
        message += f"Time: {alert_data.get('timestamp', 'N/A')}"

        return self.send_message(message, channel="#alerts")

class EmailConnector(BaseConnector):
    """Email integration for notifications and reports"""

    def __init__(self, smtp_server: str, smtp_port: int, username: str, password: str, use_tls: bool = True):
        super().__init__({
            'smtp_server': smtp_server,
            'smtp_port': smtp_port,
            'username': username,
            'password': password,
            'use_tls': use_tls
        })
        self.smtp_server = smtp_server
        self.smtp_port = smtp_port
        self.username = username
        self.password = password
        self.use_tls = use_tls

    def connect(self) -> bool:
        return self.test_connection()['success']

    def test_connection(self) -> Dict[str, Any]:
        try:
            server = smtplib.SMTP(self.smtp_server, self.smtp_port)
            if self.use_tls:
                server.starttls()
            server.login(self.username, self.password)
            server.quit()
            return {'success': True, 'message': 'Email connection successful'}
        except Exception as e:
            return {'success': False, 'error': str(e)}

    def send_email(self, to_emails: List[str], subject: str, body: str, html_body: Optional[str] = None) -> Dict[str, Any]:
        try:
            msg = MIMEMultipart('alternative')
            msg['Subject'] = subject
            msg['From'] = self.username
            msg['To'] = ', '.join(to_emails)

            # Add text part
            text_part = MIMEText(body, 'plain')
            msg.attach(text_part)

            # Add HTML part if provided
            if html_body:
                html_part = MIMEText(html_body, 'html')
                msg.attach(html_part)

            server = smtplib.SMTP(self.smtp_server, self.smtp_port)
            if self.use_tls:
                server.starttls()
            server.login(self.username, self.password)
            server.send_message(msg)
            server.quit()

            return {'success': True, 'message': f'Email sent to {len(to_emails)} recipients'}
        except Exception as e:
            return {'success': False, 'error': str(e)}

    def send_report(self, to_emails: List[str], report_data: Dict[str, Any]) -> Dict[str, Any]:
        subject = f"BMasterAI Report - {datetime.now().strftime('%Y-%m-%d %H:%M')}"

        body = f"""
BMasterAI System Report

Generated: {report_data.get('timestamp', 'N/A')}

System Health:
- Active Agents: {report_data.get('active_agents', 0)}
- Total Agents: {report_data.get('total_agents', 0)}
- CPU Usage: {report_data.get('system_metrics', {}).get('cpu', {}).get('avg', 'N/A')}%
- Memory Usage: {report_data.get('system_metrics', {}).get('memory', {}).get('avg', 'N/A')}%

Recent Alerts: {len(report_data.get('recent_alerts', []))}

Best regards,
BMasterAI System
"""

        return self.send_email(to_emails, subject, body)

class DatabaseConnector(BaseConnector):
    """Database integration for data storage and retrieval"""

    def __init__(self, db_type: str = "sqlite", connection_string: str = "bmasterai.db", **kwargs):
        super().__init__({'db_type': db_type, 'connection_string': connection_string, **kwargs})
        self.db_type = db_type
        self.connection_string = connection_string
        self.connection = None

        if db_type == "sqlite":
            self._init_sqlite()

    def _init_sqlite(self):
        """Initialize SQLite database with required tables"""
        try:
            conn = sqlite3.connect(self.connection_string)
            cursor = conn.cursor()

            # Create agents table
            cursor.execute('''
                CREATE TABLE IF NOT EXISTS agents (
                    id TEXT PRIMARY KEY,
                    name TEXT,
                    status TEXT,
                    created_at TIMESTAMP,
                    last_activity TIMESTAMP,
                    metadata TEXT
                )
            ''')

            # Create tasks table
            cursor.execute('''
                CREATE TABLE IF NOT EXISTS tasks (
                    id TEXT PRIMARY KEY,
                    agent_id TEXT,
                    name TEXT,
                    status TEXT,
                    duration_ms REAL,
                    created_at TIMESTAMP,
                    completed_at TIMESTAMP,
                    metadata TEXT,
                    FOREIGN KEY (agent_id) REFERENCES agents (id)
                )
            ''')

            # Create metrics table
            cursor.execute('''
                CREATE TABLE IF NOT EXISTS metrics (
                    id INTEGER PRIMARY KEY AUTOINCREMENT,
                    metric_name TEXT,
                    value REAL,
                    labels TEXT,
                    timestamp TIMESTAMP
                )
            ''')

            conn.commit()
            conn.close()
        except Exception as e:
            print(f"Error initializing SQLite database: {e}")

    def connect(self) -> bool:
        return self.test_connection()['success']

    def test_connection(self) -> Dict[str, Any]:
        try:
            if self.db_type == "sqlite":
                conn = sqlite3.connect(self.connection_string)
                conn.close()
                return {'success': True, 'message': 'Database connection successful'}
            else:
                return {'success': False, 'error': 'Unsupported database type'}
        except Exception as e:
            return {'success': False, 'error': str(e)}

    def store_agent_data(self, agent_id: str, name: str, status: str, metadata: Dict[str, Any] = None) -> Dict[str, Any]:
        try:
            conn = sqlite3.connect(self.connection_string)
            cursor = conn.cursor()

            cursor.execute('''
                INSERT OR REPLACE INTO agents (id, name, status, created_at, last_activity, metadata)
                VALUES (?, ?, ?, ?, ?, ?)
            ''', (
                agent_id, name, status, 
                datetime.now(timezone.utc).isoformat(),
                datetime.now(timezone.utc).isoformat(),
                json.dumps(metadata or {})
            ))

            conn.commit()
            conn.close()
            return {'success': True, 'message': 'Agent data stored successfully'}
        except Exception as e:
            return {'success': False, 'error': str(e)}

    def store_task_data(self, task_id: str, agent_id: str, name: str, status: str, 
                       duration_ms: Optional[float] = None, metadata: Dict[str, Any] = None) -> Dict[str, Any]:
        try:
            conn = sqlite3.connect(self.connection_string)
            cursor = conn.cursor()

            cursor.execute('''
                INSERT OR REPLACE INTO tasks (id, agent_id, name, status, duration_ms, created_at, completed_at, metadata)
                VALUES (?, ?, ?, ?, ?, ?, ?, ?)
            ''', (
                task_id, agent_id, name, status, duration_ms,
                datetime.now(timezone.utc).isoformat(),
                datetime.now(timezone.utc).isoformat() if status == 'completed' else None,
                json.dumps(metadata or {})
            ))

            conn.commit()
            conn.close()
            return {'success': True, 'message': 'Task data stored successfully'}
        except Exception as e:
            return {'success': False, 'error': str(e)}

    def get_agent_history(self, agent_id: str, limit: int = 100) -> List[Dict[str, Any]]:
        try:
            conn = sqlite3.connect(self.connection_string)
            cursor = conn.cursor()

            cursor.execute('''
                SELECT * FROM tasks WHERE agent_id = ? ORDER BY created_at DESC LIMIT ?
            ''', (agent_id, limit))

            columns = [description[0] for description in cursor.description]
            results = [dict(zip(columns, row)) for row in cursor.fetchall()]

            conn.close()
            return results
        except Exception as e:
            print(f"Error getting agent history: {e}")
            return []

class WebhookConnector(BaseConnector):
    """Generic webhook integration for external services"""

    def __init__(self, webhook_url: str, headers: Optional[Dict[str, str]] = None, auth_token: Optional[str] = None):
        super().__init__({'webhook_url': webhook_url, 'headers': headers, 'auth_token': auth_token})
        self.webhook_url = webhook_url
        self.headers = headers or {'Content-Type': 'application/json'}

        if auth_token:
            self.headers['Authorization'] = f'Bearer {auth_token}'

    def connect(self) -> bool:
        return self.test_connection()['success']

    def test_connection(self) -> Dict[str, Any]:
        try:
            test_payload = {'test': True, 'timestamp': datetime.now(timezone.utc).isoformat()}
            response = requests.post(self.webhook_url, json=test_payload, headers=self.headers, timeout=10)
            return {'success': response.status_code < 400, 'status_code': response.status_code}
        except Exception as e:
            return {'success': False, 'error': str(e)}

    def send_data(self, data: Dict[str, Any]) -> Dict[str, Any]:
        try:
            response = requests.post(self.webhook_url, json=data, headers=self.headers, timeout=30)
            response.raise_for_status()
            return {'success': True, 'status_code': response.status_code, 'response': response.text}
        except Exception as e:
            return {'success': False, 'error': str(e)}

class DiscordConnector(BaseConnector):
    """Discord integration for notifications"""

    def __init__(self, webhook_url: str):
        super().__init__({'webhook_url': webhook_url})
        self.webhook_url = webhook_url

    def connect(self) -> bool:
        return self.test_connection()['success']

    def test_connection(self) -> Dict[str, Any]:
        try:
            response = self.send_message("🤖 BMasterAI connection test")
            return {'success': response['success'], 'message': 'Discord connection test completed'}
        except Exception as e:
            return {'success': False, 'error': str(e)}

    def send_message(self, message: str, username: str = "BMasterAI") -> Dict[str, Any]:
        payload = {
            'content': message,
            'username': username,
            'avatar_url': 'https://cdn-icons-png.flaticon.com/512/4712/4712027.png'
        }

        try:
            response = requests.post(self.webhook_url, json=payload, timeout=10)
            response.raise_for_status()
            return {'success': True, 'response': response.text}
        except Exception as e:
            return {'success': False, 'error': str(e)}

    def send_embed(self, title: str, description: str, color: int = 0x00ff00, fields: List[Dict[str, str]] = None) -> Dict[str, Any]:
        embed = {
            'title': title,
            'description': description,
            'color': color,
            'timestamp': datetime.now(timezone.utc).isoformat(),
            'footer': {'text': 'BMasterAI'}
        }

        if fields:
            embed['fields'] = fields

        payload = {'embeds': [embed]}

        try:
            response = requests.post(self.webhook_url, json=payload, timeout=10)
            response.raise_for_status()
            return {'success': True, 'response': response.text}
        except Exception as e:
            return {'success': False, 'error': str(e)}

class TeamsConnector(BaseConnector):
    """Microsoft Teams integration"""

    def __init__(self, webhook_url: str):
        super().__init__({'webhook_url': webhook_url})
        self.webhook_url = webhook_url

    def connect(self) -> bool:
        return self.test_connection()['success']

    def test_connection(self) -> Dict[str, Any]:
        try:
            response = self.send_message("🤖 BMasterAI connection test")
            return {'success': response['success'], 'message': 'Teams connection test completed'}
        except Exception as e:
            return {'success': False, 'error': str(e)}

    def send_message(self, message: str, title: str = "BMasterAI Notification") -> Dict[str, Any]:
        payload = {
            '@type': 'MessageCard',
            '@context': 'http://schema.org/extensions',
            'themeColor': '0076D7',
            'summary': title,
            'sections': [{
                'activityTitle': title,
                'activitySubtitle': datetime.now(timezone.utc).strftime('%Y-%m-%d %H:%M:%S UTC'),
                'activityImage': 'https://cdn-icons-png.flaticon.com/512/4712/4712027.png',
                'text': message,
                'markdown': True
            }]
        }

        try:
            response = requests.post(self.webhook_url, json=payload, timeout=10)
            response.raise_for_status()
            return {'success': True, 'response': response.text}
        except Exception as e:
            return {'success': False, 'error': str(e)}

class IntegrationManager:
    """Manages all integrations and provides unified interface"""

    def __init__(self):
        self.connectors: Dict[str, BaseConnector] = {}
        self.active_integrations: List[str] = []

    def add_connector(self, name: str, connector: BaseConnector) -> Dict[str, Any]:
        try:
            if connector.connect():
                self.connectors[name] = connector
                self.active_integrations.append(name)
                return {'success': True, 'message': f'{name} connector added successfully'}
            else:
                return {'success': False, 'error': f'Failed to connect to {name}'}
        except Exception as e:
            return {'success': False, 'error': str(e)}

    def remove_connector(self, name: str) -> Dict[str, Any]:
        if name in self.connectors:
            del self.connectors[name]
            if name in self.active_integrations:
                self.active_integrations.remove(name)
            return {'success': True, 'message': f'{name} connector removed'}
        return {'success': False, 'error': f'{name} connector not found'}

    def get_connector(self, name: str) -> Optional[BaseConnector]:
        return self.connectors.get(name)

    def test_all_connections(self) -> Dict[str, Dict[str, Any]]:
        results = {}
        for name, connector in self.connectors.items():
            results[name] = connector.test_connection()
        return results

    def broadcast_message(self, message: str, exclude: List[str] = None) -> Dict[str, Dict[str, Any]]:
        exclude = exclude or []
        results = {}

        for name, connector in self.connectors.items():
            if name in exclude:
                continue

            try:
                if hasattr(connector, 'send_message'):
                    results[name] = connector.send_message(message)
                else:
                    results[name] = {'success': False, 'error': 'Connector does not support messaging'}
            except Exception as e:
                results[name] = {'success': False, 'error': str(e)}

        return results

    def send_alert_to_all(self, alert_data: Dict[str, Any]) -> Dict[str, Dict[str, Any]]:
        results = {}

        for name, connector in self.connectors.items():
            try:
                if hasattr(connector, 'send_alert'):
                    results[name] = connector.send_alert(alert_data)
                elif hasattr(connector, 'send_message'):
                    message = f"Alert: {alert_data.get('metric_name', 'Unknown')} - {alert_data.get('message', '')}"
                    results[name] = connector.send_message(message)
                else:
                    results[name] = {'success': False, 'error': 'Connector does not support alerts'}
            except Exception as e:
                results[name] = {'success': False, 'error': str(e)}

        return results

    def get_status(self) -> Dict[str, Any]:
        return {
            'total_connectors': len(self.connectors),
            'active_integrations': self.active_integrations,
            'connector_status': {name: connector.test_connection() for name, connector in self.connectors.items()}
        }

# Global integration manager
_integration_manager = None

def get_integration_manager() -> IntegrationManager:
    global _integration_manager
    if _integration_manager is None:
        _integration_manager = IntegrationManager()
    return _integration_manager



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Union, Iterable, Optional
from typing_extensions import Literal, Required, TypeAlias, TypedDict

from .grader_inputs_param import GraderInputsParam
from ..shared.reasoning_effort import ReasoningEffort
from ..responses.response_input_text_param import ResponseInputTextParam
from ..responses.response_input_audio_param import ResponseInputAudioParam

__all__ = [
    "ScoreModelGraderParam",
    "Input",
    "InputContent",
    "InputContentOutputText",
    "InputContentInputImage",
    "SamplingParams",
]


class InputContentOutputText(TypedDict, total=False):
    """A text output from the model."""

    text: Required[str]
    """The text output from the model."""

    type: Required[Literal["output_text"]]
    """The type of the output text. Always `output_text`."""


class InputContentInputImage(TypedDict, total=False):
    """An image input block used within EvalItem content arrays."""

    image_url: Required[str]
    """The URL of the image input."""

    type: Required[Literal["input_image"]]
    """The type of the image input. Always `input_image`."""

    detail: str
    """The detail level of the image to be sent to the model.

    One of `high`, `low`, or `auto`. Defaults to `auto`.
    """


InputContent: TypeAlias = Union[
    str,
    ResponseInputTextParam,
    InputContentOutputText,
    InputContentInputImage,
    ResponseInputAudioParam,
    GraderInputsParam,
]


class Input(TypedDict, total=False):
    """
    A message input to the model with a role indicating instruction following
    hierarchy. Instructions given with the `developer` or `system` role take
    precedence over instructions given with the `user` role. Messages with the
    `assistant` role are presumed to have been generated by the model in previous
    interactions.
    """

    content: Required[InputContent]
    """Inputs to the model - can contain template strings.

    Supports text, output text, input images, and input audio, either as a single
    item or an array of items.
    """

    role: Required[Literal["user", "assistant", "system", "developer"]]
    """The role of the message input.

    One of `user`, `assistant`, `system`, or `developer`.
    """

    type: Literal["message"]
    """The type of the message input. Always `message`."""


class SamplingParams(TypedDict, total=False):
    """The sampling parameters for the model."""

    max_completions_tokens: Optional[int]
    """The maximum number of tokens the grader model may generate in its response."""

    reasoning_effort: Optional[ReasoningEffort]
    """
    Constrains effort on reasoning for
    [reasoning models](https://platform.openai.com/docs/guides/reasoning). Currently
    supported values are `none`, `minimal`, `low`, `medium`, `high`, and `xhigh`.
    Reducing reasoning effort can result in faster responses and fewer tokens used
    on reasoning in a response.

    - `gpt-5.1` defaults to `none`, which does not perform reasoning. The supported
      reasoning values for `gpt-5.1` are `none`, `low`, `medium`, and `high`. Tool
      calls are supported for all reasoning values in gpt-5.1.
    - All models before `gpt-5.1` default to `medium` reasoning effort, and do not
      support `none`.
    - The `gpt-5-pro` model defaults to (and only supports) `high` reasoning effort.
    - `xhigh` is supported for all models after `gpt-5.1-codex-max`.
    """

    seed: Optional[int]
    """A seed value to initialize the randomness, during sampling."""

    temperature: Optional[float]
    """A higher temperature increases randomness in the outputs."""

    top_p: Optional[float]
    """An alternative to temperature for nucleus sampling; 1.0 includes all tokens."""


class ScoreModelGraderParam(TypedDict, total=False):
    """A ScoreModelGrader object that uses a model to assign a score to the input."""

    input: Required[Iterable[Input]]
    """The input messages evaluated by the grader.

    Supports text, output text, input image, and input audio content blocks, and may
    include template strings.
    """

    model: Required[str]
    """The model to use for the evaluation."""

    name: Required[str]
    """The name of the grader."""

    type: Required[Literal["score_model"]]
    """The object type, which is always `score_model`."""

    range: Iterable[float]
    """The range of the score. Defaults to `[0, 1]`."""

    sampling_params: SamplingParams
    """The sampling parameters for the model."""



# -*- coding: utf-8 -*-

# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Utility functions for all pre-evaluation steps."""

from __future__ import annotations

from concurrent import futures
from typing import Callable, Optional, Set, TYPE_CHECKING, Union, List

from google.cloud.aiplatform import base
from google.cloud.aiplatform_v1beta1.types import (
    content as gapic_content_types,
)
from vertexai import generative_models
from vertexai.preview.evaluation import _base as evaluation_base
from vertexai.preview.evaluation import constants
from vertexai.preview.evaluation import multimodal_utils
from vertexai.preview.evaluation import (
    prompt_template as prompt_template_base,
)


if TYPE_CHECKING:
    import pandas as pd

try:
    from tqdm import tqdm
except ImportError:
    raise ImportError(
        'tqdm is not installed. Please install the SDK using "pip install'
        ' google-cloud-aiplatform[evaluation]"'
    )


_LOGGER = base.Logger(__name__)
_SUCCESSFUL_FINISH_REASONS = [
    gapic_content_types.Candidate.FinishReason.STOP,
    gapic_content_types.Candidate.FinishReason.MAX_TOKENS,
    # Many responses have this finish reason
    gapic_content_types.Candidate.FinishReason.FINISH_REASON_UNSPECIFIED,
]


def _assemble_prompt(
    row: "pd.Series",
    prompt_template: Union[prompt_template_base.PromptTemplate, str],
) -> str:
    """Assembles the prompt template with the given row data."""
    if isinstance(prompt_template, str):
        prompt_template = prompt_template_base.PromptTemplate(prompt_template)
    _check_variable_columns_exist(row, prompt_template.variables)
    return str(
        prompt_template.assemble(
            **row[list(prompt_template.variables)].astype(str).to_dict()
        )
    )


def _generate_content_text_response(
    model: generative_models.GenerativeModel, prompt: str, max_attempts: int = 3
) -> str:
    """Generates a text response from Gemini model from a text prompt with retries .

    Args:
        model: The Gemini model instance.
        prompt: The prompt to send to the model.
        max_attempts: Maximum number of attempts for response generation.

    Returns:
        The text response from the model.

    Raises:
        RuntimeError if the prompt or the response for the prompt is blocked for
        safety reasons.
    """
    for attempt in range(max_attempts):
        try:
            response = model.generate_content(prompt)
            if not response.candidates:
                error_message = (
                    f"The model response was blocked due to"
                    f" {response._raw_response.prompt_feedback.block_reason.name}.\n"
                    f"Blocked reason message:"
                    f" {response._raw_response.prompt_feedback.block_reason_message}.\n"
                    "The input prompt may be blocked for safety reasons.\n"
                    f"Prompt: {prompt}.\n"
                    f"Attempt: {attempt + 1}/{max_attempts}"
                )
                _LOGGER.warning(error_message)
                break
            else:
                candidate = response.candidates[0]
                if candidate.finish_reason not in _SUCCESSFUL_FINISH_REASONS:
                    error_message = (
                        "The model response did not finish"
                        " successfully.\n"
                        f"Finish reason: {candidate.finish_reason}.\n"
                        f"Finish message: {candidate.finish_message}.\n"
                        f"Safety ratings: {candidate.safety_ratings}.\n"
                        "Please adjust the model safety_settings, or"
                        " try a different prompt.\n"
                        f"Attempt: {attempt + 1}/{max_attempts}"
                    )
                    _LOGGER.warning(error_message)
                else:
                    return response.candidates[0].content.parts[0].text
        except Exception as e:
            error_message = (
                f"Failed to generate response candidates from Gemini model"
                f" {model._model_name}.\n"
                f"Error: {e}.\n"
                f"Prompt: {prompt}.\n"
                f"Attempt: {attempt + 1}/{max_attempts}"
            )
            _LOGGER.warning(error_message)
        if attempt < max_attempts - 1:
            _LOGGER.info(
                f"Retrying response generation for prompt: {prompt}, attempt"
                f" {attempt + 1}/{max_attempts}..."
            )

    final_error_message = (
        f"Failed to generate response from Gemini model {model._model_name}.\n"
        f"Prompt: {prompt}."
    )
    _LOGGER.error(final_error_message)
    return constants.RESPONSE_ERROR


def _generate_responses_from_gemini_model(
    model: generative_models.GenerativeModel,
    df: "pd.DataFrame",
    rubric_generation_prompt_template: Optional[str] = None,
) -> List[str]:
    """Generates responses from Gemini model for the given evaluation dataset.

    Args:
        model: The Gemini model instance.
        df: Evaluation Dataset.
    Returns:
        The list of model responses.
    """
    _LOGGER.info(
        f"Generating a total of {df.shape[0]} "
        f"responses from Gemini model {model._model_name.split('/')[-1]}."
    )
    tasks = []
    with tqdm(total=len(df)) as pbar:
        with futures.ThreadPoolExecutor(max_workers=constants.MAX_WORKERS) as executor:
            for idx, row in df.iterrows():
                if rubric_generation_prompt_template:
                    input_columns = prompt_template_base.PromptTemplate(
                        rubric_generation_prompt_template
                    ).variables
                    if multimodal_utils.is_multimodal_instance(
                        row[list(input_columns)].to_dict()
                    ):
                        prompt = multimodal_utils._assemble_multi_modal_prompt(
                            rubric_generation_prompt_template, row, idx, input_columns
                        )
                    else:
                        prompt = _assemble_prompt(
                            row, rubric_generation_prompt_template
                        )
                else:
                    prompt = row[constants.Dataset.PROMPT_COLUMN]
                task = executor.submit(
                    _generate_content_text_response,
                    prompt=prompt,
                    model=model,
                )
                task.add_done_callback(lambda _: pbar.update(1))
                tasks.append(task)
        responses = [future.result() for future in tasks]
    return responses


def _generate_response_from_custom_model_fn(
    model_fn: Callable[[str], str], eval_dataset: "pd.DataFrame"
) -> List[str]:
    """Generates responses from a custom model function.

    Args:
        model_fn: The custom model function.
        eval_dataset: Evaluation Dataset.
    Returns:
        The list of model responses.
    """
    max_workers = 5

    _LOGGER.info(
        f"Generating a total of {eval_dataset.shape[0]} "
        "responses from the custom model function."
    )
    tasks = []
    try:
        with tqdm(total=len(eval_dataset)) as pbar:
            with futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
                for _, row in eval_dataset.iterrows():
                    task = executor.submit(
                        model_fn, row[constants.Dataset.PROMPT_COLUMN]
                    )
                    task.add_done_callback(lambda _: pbar.update(1))
                    tasks.append(task)
    except (ValueError, IndexError) as e:
        _LOGGER.warning(f"Failed to generate response from model function: {e}")

    responses = [task.result() for task in tasks]
    return responses


def populate_eval_dataset_with_model_responses(
    responses: List[str],
    evaluation_run_config: evaluation_base.EvaluationRunConfig,
    is_baseline_model: bool = False,
) -> None:
    """Populates the evaluation dataset with model responses.

    Args:
        responses: The list of model responses.
        evaluation_run_config: Evaluation Run Configurations.
        is_baseline_model: Whether the model is a baseline model for
          PairwiseMetric.
    """
    df = evaluation_run_config.dataset.copy()
    if is_baseline_model:
        evaluation_run_config.dataset = df.assign(baseline_model_response=responses)
    else:
        evaluation_run_config.dataset = df.assign(response=responses)

    _LOGGER.info(
        f"All {evaluation_run_config.dataset.shape[0]} responses are successfully"
        f" generated from model."
    )


def _check_variable_columns_exist(
    dataset_row: "pd.Series", variable_names_set: Set[str]
) -> None:
    """Checks if all variable names exist in the dataset columns.

    Args:
        dataset: The dataset to evaluate.
        variable_names_set: A set of variable names.

    Raises:
        ValueError: If any variable names do not exist in the dataset columns
        or the prompt template is invalid.
    """
    actual_column_names_set = set(dataset_row.to_dict().keys())
    if not variable_names_set.issubset(actual_column_names_set):
        missing_columns = variable_names_set - actual_column_names_set
        raise ValueError(
            "Failed to assemble prompt template: The following column(s) are"
            f" missing: {', '.join(missing_columns)}. "
            f"Please verify prompt_template variables {variable_names_set} and "
            f"evaluation dataset column names {actual_column_names_set}."
        )
